#pragma TextEncoding = "UTF-8"
#pragma rtGlobals=3				// Use modern global access method and strict Wave access.
#pragma DefaultTab={3,20,4}		// Set default tab width in Igor Pro 9 and later
#pragma IgorVersion=9.0.1		// 38365
#pragma version=150				// version 1.5.0

//___________________________________________________________________________________________________	
//
//	QLA netCDF reader - Quacks Like A netCDF
//	A tool for loading netCDF-3 and netCDF-4 files consistently into Igor Pro 
//___________________________________________________________________________________________________	
//
//	Copyright ©2022
//	Chris Reed (FAAM Airborne Laboratory, Building 146, Central Avenue, Cranfield University, Bedfordshire, MK43 0AL, UK)
//	chris.reed@faam.ac.uk
//___________________________________________________________________________________________________	
//
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
//___________________________________________________________________________________________________	

//#define AlphaNumSort 									// sort attributes alphanumerically, instead of creation order
//#define CharAsText									// convert char data to 1D text waves 

// modifiers for netCDF names which might conflict with igor functions
Static StrConstant timeWvStrConstant = "Time_UTC"		// constant to name all 'time/Time/TIME' waves loaded from netCDF
Static StrConstant illegalPreStr = ""					// prefix to add to any netCDF variable with a name that conflicts with an Igor function	
Static StrConstant illegalSuffStr = ""					// suffix to add to any netCDF variable with a name that conflicts with an Igor function

// name for metadata waves which may be created
Static StrConstant globAttWvStr = "global_attributes"	// name of global attributes wave
Static StrConstant grpAttWvStr	= "group_attributes"	// name of group attributes wave
Static StrConstant fileAttWvStr = "file_attributes"		// name of file attributes wave
Static StrConstant dimWvStr = "dimensions"				// name of dimensions wave

// fill text for errors or null attributes
Static StrConstant	attErrorStr = "_ERROR_"				// text to display when an error was encountered loading an attribute
Static StrConstant	attNullStr = "_NULL_"				// text to display when an null attribute was encountered

//___________________________________________________________________________________________________	

// HDF5/netCDF-4 constants, DO NOT MODIFY
Static Constant padValue = 0x00							// padding character for CHAR datasets
Static StrConstant HDF5markup = "DIMENSION_LIST;DIMENSION_SCALE;CLASS;NAME;REFERENCE_LIST;"
Static StrConstant NCmarkup = "_nc3_strict;_NCProperties;_Netcdf4Coordinates;_Netcdf4Dimid;"

// file extensions that can be supported
Static StrConstant fileFilters = "netCDF Data Files (*.nc,*.cdf):.nc,.cdf;HDF5 Data Files (*.h5,*.hdf5):.h5,.hdf5;"

//___________________________________________________________________________________________________	
// menu
Menu "QLA netCDF" 
	"Load selected",/Q,Print "QLA netCDF: \r\n\t"+ReplaceString(";",QLA_netCDF(2),"\r\n\t")
	"\\M0Load w/o unpacking",/Q,Print "QLA netCDF: \r\n\t"+ReplaceString(";",QLA_netCDF(2,unpack=4),"\r\n\t")
	"Load and transpose",/Q,Print "QLA netCDF: \r\n\t"+ReplaceString(";",QLA_netCDF(2,trans=1),"\r\n\t")
	"\\M0Load w/dump",/Q,Print "QLA netCDF: \r\n\t"+ReplaceString(";",QLA_netCDF(2,dump=1),"\r\n\t")
	"-"
	"View dump",/Q,QLA_dump()
	"-"
	"What is it?",/Q,Print QLA_kind("")
End
//___________________________________________________________________________________________________	
// interactivly load multiple netCDF3/4s or silently load a single netCDF if called from within another function
// options - where data is loaded 
// 		0 - load netCDFs in root
// 		1 - load netCDFs in current data folder 
// 		2 - load netCDFs in folder relative to root 
// 		3 - load netCDFs in folder relative to current data folder 
// unpack - mode for unpacking data
//		0 - default - Range check and data scaling is enabled, fill data is set to NaN, unpack=0 or ommit
//		1 - Range check and data scaling is enabled, unpack=1 
// 		2 - Range check is enabled, data scaling is disabled, unpack=2
// 		3 - Range check is disabled, data scaling is enabled, unpack=3
// 		4 - Range check and data scaling is disabled, unpack=4
// trans - transpose - transpose 2D datasets along the long axis, trans=1 
// dump - display a dump of the netCDF header in a notebook, dump=1  
Function/S QLA_netCDF(int DFoption[,int unpack,int transpose,int dump])
 
 	// check for the presence of the netCDF xop
 	#if !GrepString(IgorInfo(10),"NetCDF")
		QLA_error()
	#endif
 
	String savedDF = GetDataFolder(1)
	String fNameStr,dataFolderStr,returnStr=""

	// get a dialogue to select files
	Open /D /R /MULT=1 /F=fileFilters /M="Select netCDFs to open" fileID
	
	// handle if no files are chosen
	if (strlen(S_fileName) == 0)
		return "" // user didn't choose a file
	endif

	// loop through the list of files
	for (fNameStr : ListToTextWave(S_filename,"\r"))
	
		// get a legal name for the possible data folder
		dataFolderStr = PossiblyQuoteName(CleanUpName(ParseFilePath(3,fNameStr,":",0,0),1))
		
		// get a list of options to load the data to
		Make /O /FREE /T=4 dataFolderWv = {"root:",savedDF,"root:" + dataFolderStr,savedDF + dataFolderStr}
		
		// make a new data folder if it doesn't already exist and switch to it
		if (CmpStr(dataFolderWv[DFoption],savedDF,2))
			NewDataFolder /O /S $dataFolderWv[DFoption]
		endif
		
		// load the file and collect the output
		returnStr = AddListItem(QLA_get_nc(fNameStr,unpack,transpose),returnStr,";",Inf) // load the file
		
		// if nothing was loaded kill any empty data folder 
		if ((DFoption == 2 || DFoption == 3) && (!CountObjects("",1)))
			KillDataFolder /Z GetDataFolderDFR() // kill the  current DF
		elseif (dump) 
			QLA_dump() // optionally call QLA_dump if requested
		endif
	endfor

	SetDataFolder savedDF // return to the start
	
	return returnStr
End
//___________________________________________________________________________________________________	
// function to load netCDF3/4 & hdf5 files, metadata is parsed into the note of the variable. 
// Group and global attribute get their own text wave to attach metadata to
// fill value applied, range checking applied optionally with unpack options (0-4), 
// default is 0 - range check and fill value applied, transposition of multidimensional datasets optionally
Function/S QLA_get_nc(String fNameStr,int mode,int transpose)
	
	// get the type of file
	int type = netCDFtype(fNameStr),error ; String varNameStr,errorStr=""
		
	// load the file
	switch (type)
		case 3:	// netCDF-3 classic
			[error,errorStr,varNameStr] = getNC3(fNameStr,mode,transpose)
			break
		case 4:	// netCDF-4
			[error,errorStr,varNameStr] = getNC4(fNameStr,mode,transpose)
			break
		case 5: // CDF5
			return "CDF-5 files are not currently supported: " + ParseFilePath(0,fNameStr,":",1,0)
		case 6: // netCDF-3 64-bit offset
			[error,errorStr,varNameStr] = getNC3(fNameStr,mode,transpose)
			break
		case 7: // netCDF-4 classic
			[error,errorStr,varNameStr] = getNC4(fNameStr,mode,transpose)
			break
		default:
			return "Unknown file format: " + ParseFilePath(0,fNameStr,":",1,0)
	endswitch

	// get file information
	GetFileFolderInfo /Q /Z fNameStr ; error = (error || V_flag) ? 1 : 0
	
	// make file information
	Make /O /T /N=0 $fileAttWvStr/Wave=fileInfo_Wv
	Note /k fileInfo_Wv, "// file attributes:\r"
	QLA_set_att(fileInfo_Wv,"filename","\"" + ParseFilePath(0,S_path,":",1,0) + "\";")
	QLA_set_att(fileInfo_Wv,"path","\"" + ParseFilePath(1,S_path,":",1,0) + "\";")
	QLA_set_att(fileInfo_Wv,"type","\"" + netCDFtypeStr(type)+"\";")
	QLA_set_att(fileInfo_Wv,"size","\"" + num2str(V_logEOF/(1024^2),"%.3f mb\";"))
	QLA_set_att(fileInfo_Wv,"created","\"" + Secs2Date(V_modificationDate,2) + ", " + Secs2Time(V_modificationDate,3) + "\";")
	QLA_set_att(fileInfo_Wv,"variables",num2str(ItemsInList(varNameStr),"%d; // int"))
	QLA_set_att(fileInfo_Wv,"names","\"" + varNameStr + "\";")

	// was there an error 
	errorStr = SelectString(error,fNameStr,"Errors occured while loading: " + fNameStr + ReplaceString("\r",errorStr,"\r\n\t"))

	return errorStr  
End
//___________________________________________________________________________________________________
// return the dimesnion names from the CDL syntax
Function/S QLA_get_dims2(Wave dataWv)
	
	String rtnStr="", dimStr,dataStr,matchStr ; int i
	
	dataStr = ParseFilePath(0,note(dataWv),";",0,0)
	matchStr = ".*"+NameOfWave(dataWv) + "\("

	for (i=0;i<WaveDims(dataWv);i++)
		SplitString /E=(matchStr + "(.*?)=.*") dataStr,dimStr
		matchStr += dimStr + "=.*, "
		rtnStr += dimStr + ";"
	endfor
	
	return rtnStr
end 
//___________________________________________________________________________________________________	
// return the dimesnion names	 from the labels
Function/S QLA_get_dims(Wave dataWv)
	
	String rtnStr="" ; int i
	
	for (i=0;i<WaveDims(dataWv);i++)
		rtnStr += GetDimLabel(dataWv,i,-1) + ";"
	endfor
	
	return rtnStr
End 
//___________________________________________________________________________________________________	
// set or modify the named attribute
Function QLA_set_att(Wave Wv, String attKey, String attVal)
	
	Note /K Wv, SelectString(cmpstr(attVal,"",2),RemoveByKey("\t:" + attKey,note(Wv)," = ","\r"),ReplaceStringByKey("\t:" + attKey,note(Wv),attVal," = ","\r"))
	
	return 0
End
//___________________________________________________________________________________________________	
// get the one of the named attbributes as a string
Function/S QLA_get_att(Wave Wv,String attKeys)

	String rtnStr="", attKey
  	
  	for (attKey : ListToTextWave(attKeys,";"))
  		SplitString/E=("(.*?);.*") ReplaceString("\"",StringByKey("\t:"+attKey,Note('Wv')," = ","\r"),""),rtnStr
  		if (strlen(rtnStr)>0)
  			break
  		endif
  	endfor 
  	
  	return rtnStr
End
//___________________________________________________________________________________________________	
// list the attributes
Function/S QLA_inq_att(Wave Wv)

	String attStr,valStr,lstStr="",rtnStr=""

	for (lstStr : ListToTextWave(note(Wv),"\r\t:"))
		lstStr = ReplaceString("\"",lstStr,"")
		SplitString /E=("(.*?) = (.*?)") lstStr,attStr,valStr 
		if (strlen(attStr) > 0)
			rtnStr = AddListItem(AttStr,rtnStr,";",Inf)
		endif
	endfor
	
	return rtnStr
End
//___________________________________________________________________________________________________		
// apply unpack preferences 
Static Function unpackNC(Wave dataWv,int mode)

	// list of all the attributes associated with dataWv
	String attsLst = QLA_inq_att(dataWv)
	
	double fillVal ; String fillStr
	
	//	// is unpack dissabled
	if (mode == 4)
		return 0
	endif
	
	// units
	if (GrepString(attsLst,"units"))
		SetScale d 0,0, QLA_get_att(dataWv,"units"), dataWv
	endif
	
	// get a fill value 
	if (GrepString(attsLst,"_FillValue") && (WaveType(dataWv,1) == 1) && (WaveType(dataWv) < 8) && (numpnts(dataWv) > 0))
		Wave attWv = varLst2wave(QLA_get_att(dataWv,"_FillValue"),",",WaveType(dataWv))
		fillVal = attWv[inf]
	elseif (GrepString(attsLst,"missing_value") && (WaveType(dataWv,1) == 1) && (WaveType(dataWv) < 8) && (numpnts(dataWv) > 0))
		Wave attWv = varLst2wave(QLA_get_att(dataWv,"missing_value"),",",WaveType(dataWv))
		fillVal = attWv[inf]
	endif
		
	// scale_factor - before offset
	if ((mode < 2 || mode == 3) && (GrepString(attsLst,"scale_factor")) && (WaveType(dataWv,1) == 1) && (numpnts(dataWv) > 0))
		Wave attWv = varLst2wave(QLA_get_att(dataWv,"scale_factor"),",",WaveType(dataWv))
		if (attWv[Inf] != 1)	// waste of time multiplying every point by 1
			Redimension /S dataWv
			MultiThread dataWv *= attWv[Inf]
		endif
	endif

	// add_offset - after scaling
	if ((mode < 2 || mode == 3) && (GrepString(attsLst,"add_offset")) && (WaveType(dataWv,1) == 1) && (numpnts(dataWv) > 0))
		Wave attWv = varLst2wave(QLA_get_att(dataWv,"add_offset"),",",WaveType(dataWv))
		if (attWv[Inf] != 0)	// waste of time adding zero to every point
			MultiThread dataWv += attWv[Inf]
		endif
	endif				
	
	// valid_range	
	if ((mode < 3) && (GrepString(attsLst,"valid_range")) && (WaveType(dataWv,1) == 1) && (numpnts(dataWv) > 0))
		Wave attWv = varLst2wave(QLA_get_att(dataWv,"valid_range"),",",WaveType(dataWv))						
		MultiThread dataWv = (attWv[0] > dataWv[p][q][r][s])||(dataWv[p][q][r][s] > attWv[Inf]) ? fillVal : dataWv[p][q][r][s]
	endif						
	
	// valid_max
	if ((mode < 3) && (GrepString(attsLst,"valid_max")) && (WaveType(dataWv,1) == 1) && (numpnts(dataWv) > 0))
		Wave attWv = varLst2wave(QLA_get_att(dataWv,"valid_max"),",",WaveType(dataWv))
		MultiThread dataWv = (dataWv[p][q][r][s] > attWv[Inf]) ? fillVal : dataWv[p][q][r][s]
	endif
	
	// valid_min
	if ((mode < 3) && (GrepString(attsLst,"valid_min")) && (WaveType(dataWv,1) == 1) && (numpnts(dataWv) > 0))
		Wave attWv = varLst2wave(QLA_get_att(dataWv,"valid_min"),",",WaveType(dataWv))
		MultiThread dataWv = (dataWv[p][q][r][s] < attWv[Inf]) ? fillVal : dataWv[p][q][r][s]
	endif
	
	// _FillValue - set to NaN in floating point waves
	if ((mode == 0) && (GrepString(attsLst,"_FillValue")) && (WaveType(dataWv,1) == 1) && (WaveType(dataWv) < 8) && (numpnts(dataWv) > 0))
		Wave attWv = varLst2wave(QLA_get_att(dataWv,"_FillValue"),",",WaveType(dataWv))
		MultiThread dataWv = (dataWv[p][q][r][s] == attWv[Inf]) ? NaN : dataWv[p][q][r][s]
	elseif ((mode == 0) && (GrepString(attsLst,"_FillValue")) && (WaveType(dataWv,1) == 2) && (numpnts(dataWv) > 0))
		Wave/T textWv = dataWv ; fillStr = QLA_get_att(dataWv,"_FillValue")
		MultiThread textWv = ReplaceString(fillStr,textWv,"") // is this even allowed / are _FillValues allowed in string/char data?
	endif		 			
	
	// missing_value	 - set to NaN in floating point waves			 			
	if ((mode == 0) && (GrepString(attsLst,"missing_value")) && (WaveType(dataWv,1)==1) && (WaveType(dataWv) < 8) && (numpnts(dataWv) > 0))
		Wave attWv = varLst2wave(QLA_get_att(dataWv,"missing_value"),",",WaveType(dataWv))
		MultiThread dataWv = (dataWv[p][q][r][s] == attWv[Inf]) ? NaN : dataWv[p][q][r][s]
	elseif ((mode == 0) && (GrepString(attsLst,"missing_value")) && (WaveType(dataWv,1) == 2) && (numpnts(dataWv) > 0))
		Wave/T textWv = dataWv ; fillStr = QLA_get_att(dataWv,"missing_value")
		MultiThread textWv = ReplaceString(fillStr,textWv,"") // is this even allowed / are missing_values allowed in string/char data?
	endif
	
	return 0		 	
End
//___________________________________________________________________________________________________	
// function to get a list of netCDF/HDF5 dimensions from DIMENSION_LIST attribute
Static Function/S getHDF5Dimensions(double groupID, String WvStr)

	DFREF dfrSave = GetDataFolderDFR()

	SetDataFolder NewFreeDataFolder()	// Create new free data folder

	String rtnStr="",dimPathStr,dimWvStr
	
	HDF5ListAttributes /Type=2 groupID, WvStr
	
	// variabe has dimensions attached
	if (GrepString(S_HDF5ListAttributes,"DIMENSION_LIST"))
	
		// load the dimension list attribute 
		HDF5LoadData /O /A="DIMENSION_LIST" /TYPE=2 /N=temp_DIMENSION_LIST /Q /Z groupID, WvStr
		
		// get a list of all possible dimension list waves
		String dimWvLst = WaveList("temp_DIMENSION_LIST*",";","")
		if (V_flag == 0)	
			
			for (dimWvStr : ListToTextWave(dimWvLst,";"))
				Wave /T temp_DimLstWv = $dimWvStr
				rtnStr = AddListItem(ParseFilePath(0,temp_DimLstWV[0],"/",1,0),rtnStr,";",Inf)
			endfor
		endif
	endif
	
	// variable is a dimension
	HDF5DimensionScale dataset={groupID,WvStr}, isScale
	if (V_Value)
		rtnStr = ParseFilePath(0,WvStr,"/",1,0) + ";"
		
		// if a dimension has dimensions attached test for thier names - special case for char
		// as of netCDF 4.8.1 _Netcdf4Coordinates is everywhere - possible bug in that version so catch here.
		if (GrepString(S_HDF5ListAttributes,"_Netcdf4Coordinates") && (typeOnDisk(WvStr, groupID) == 1))
		
			// load the dimension list attribute 
			HDF5LoadData /O /A="_Netcdf4Coordinates" /TYPE=2 /N=temp_Netcdf4Coordinates /Q /Z groupID, WvStr
			if (V_flag == 0) 
				Wave temp_Netcdf4Coordinates	// catch possible error
				
				// get a list of all the variables in the file
				HDF5ListGroup /R groupID, "/"

				// loop though all the variables in the file
				for (dimPathStr : ListToTextWave(S_HDF5ListGroup,";"))
			
					// attempt to load the _Netcdf4Dimid attribute 
					HDF5LoadData /Q /Z /O /A="_Netcdf4Dimid" /TYPE=2 /N=tempWv ,groupID ,dimPathStr
					
					// if the _Netcdf4Dimid attribute exists 
					if (V_flag == 0) 
						Wave tempWv
						
						// find the dimension that matches the last (second) value
						if (tempWv[inf] == temp_Netcdf4Coordinates[inf])
							rtnStr += ParseFilePath(0,dimPathStr,"/",1,0) + ";"
							break
						endif
					endif	
				endfor 
			endif
		endif	
	endif
	
	SetDataFolder dfrSave // killing the free data folder
	
	return rtnStr
End	
//___________________________________________________________________________________________________	
// check if the dataset is a dimension even if it isn't attached to a variable, return its size
Static Function/S isHDF5Dimension(double groupID,String WvStr)

	double dummy ; String rtnStr=""
	
	HDF5DimensionScale dataset={groupID,WvStr}, isScale

	if (V_Value)
 
 		STRUCT HDF5DataInfo di	
		InitHDF5DataInfo(di)
		dummy = HDF5DatasetInfo(groupID, WvStr, 0, di)
		WvStr = ParseFilePath(0,WvStr,"/",1,0)
		
		rtnStr = "\t" + Wvstr + SelectString((di.maxdims[0] == -1),num2str(di.dims[0]," = %d;\r"),num2str(di.dims[0]," = UNLIMITED;   // (%d currently)\r"))
		
	endif

	return rtnStr
End
//___________________________________________________________________________________________________	
// function to return a formatted string with netCDF dependencies in it 
// also set dim labels with those dependencies if dimensions has >0 size
Static Function setDependencies(Wave Wv, String dimLstStr)

	int i ; String dimStr = ""
	String rtnStr = TypeToInfo(WaveType(Wv) + isChar(Wv)) + " " + QLA_get_att(Wv,"short_name") + "("

	for (dimStr : ListToTextWave(dimLstStr,";"))
		if ((DimSize(wv,i) > 0) && (strlen(dimLstStr) > 0)) // cannot set a dim label on a 0 size dimension
			SetDimLabel i, -1, $cleanUpDimLabel(dimStr), Wv
		endif
		rtnStr += dimStr + num2str(DimSize(Wv,i),"=%d, ")
		i++
	endfor 
	
	Note/K Wv, RemoveEnding(rtnStr,", ")+");\r" + note(Wv)
	
	// deal with scalars
	Note/K Wv, ReplaceString("()",note(Wv),"")
	
	return 0
End
//___________________________________________________________________________________________________	
// Function to determine the type of netCDF variable
Static Function/S TypeToInfo(int type)
	
	String info = ""

	switch (type)
		case 0: // added CR 04/20
			info = "string"
			break
		case 1: // added CR 05/21
			info = "char"
			break
		case 2:
			info = "float" // real
			break
		case 4:
			info = "double"
			break
		case 8:
			info = "byte"
			break
		case 16:
			info = "short"
			break
		case 32:
			info = "int" // long
			break
		case 72:  // added CR 04/20
			info = "ubyte"
			break
		case 80:  // added CR 04/20
			info = "ushort"
			break
		case 96:  // added CR 04/20
			info = "uint"	// ulong
			break
		case 128:  // added CR 02/21
			info = "int64"	// llong
			break
		case 192:  // added CR 02/21
			info = "uint64"	// ullong
			break
	endswitch

	return info
End
//___________________________________________________________________________________________________  
// return a path possibly quoted for liberal names
Static Function/S PossiblyQuotePath(String pathStr, String sepStr)
	
	String rtnStr="",subStr=""
	
	for (subStr : ListToTextWave(pathStr + sepStr,sepStr))
		if (numtype(strlen(subStr)) == 0)
			rtnStr = AddListItem(PossiblyQuoteName(subStr),rtnStr,sepStr,Inf)
		else
			rtnStr = AddListItem("",rtnStr,sepStr,Inf)	
		endif
	endfor
	
	return RemoveEnding(rtnStr,sepStr)
End
//___________________________________________________________________________________________________	
// convert a stringlist into a wave of given type in formatstring 
Static Function/WAVE varLst2wave(String varLstStr,String sepStr,int type)

	int n = ItemsInList(varLstStr, sepStr)
	Make /O /N=(n) /Y=(type) /FREE outWv = str2num(StringFromList(p,varLstStr, sepStr))

	return outWv
End
//___________________________________________________________________________________________________	
// control printf or maybe chris printf!
Static Function/S cprintf(Wave Wv)

	String rtnStr="",cmdStr ; int i,j,type = WaveType(Wv) 
	
	// loop though individual values in wave - inefficient but attributes won't have more than a few values
	for (double var : Wv)
		switch (type)
			case 2:		// single precision
				[i,j] = getPrecision(var,7) // single can represent ~ 7 dp
				sprintf cmdStr,"%#*.*f", i,j,var
				break
			case 4:		// double precision
				[i,j] = getPrecision(var,15) // double can represent ~ 15 dp
				sprintf cmdStr,"%#*.*f", i,j,var
				break
			default:	// integer types
				sprintf cmdStr, "%d", var
		endswitch
		rtnStr = AddListItem(cmdStr,rtnStr,", ",Inf)
	endfor 

	return RemoveEnding(rtnStr,", ")
End
//___________________________________________________________________________________________________	
// get the scale and precision of a number, i.e. number of decimal places 
// number of DP based on how large the integer part is to avoid overstated precision
Static Function [int integers,int decimals] getPrecision(double num, int exponent)
	
	// catch Inf and NaN values
	if (numtype(num) > 0)
		return [0,0]
	endif
	
	// initialize the outputs
	decimals = 0
	integers = ceil(log(abs(num)))   
	
	double threshold = 5*10^-(exponent - integers + 1) 
	num -= floor(num)	
	
	do
		num *= 10.0
		threshold *= 10.0
		decimals += 1
	while (abs(num - round(num)) > threshold) 
	
	return [integers,decimals] 
End
//___________________________________________________________________________________________________	
// function called if the netCDF XOP isn't found
Static Function QLA_error()
	DoAlert /T="QLA netCDF loader" 0,"Please install the NetCDF XOP and restart Igor"
	
	// try and give some help for the netCDF XOP
	DisplayHelpTopic /Z "Installing NetCDF XOP"
	if (V_flag != 0)
		// help for XOPs in general
		DisplayHelpTopic /Z "File Loader XOPs"
	endif
	Abort
	
	return 0
End
//___________________________________________________________________________________________________
// get the dimension imformation from a netCDF-3
Static Function/S getNCdimensionStr(double fileID)

	String rtnStr = "dimensions:\r",varStr
	
	// get unlimited dimensions
	NC_ListObjects /Z /Q /T="udim" fileID 
	if (V_flag == 0)
		varStr = RemoveEnding(S_names) // only 1 unlimited dimension is allowed in netCDF classic
		NC_Inquire /Z fileID, varStr
		rtnStr = AddListItem("\t" + varStr + " = " + SelectString(V_flag == 0,"UNKNOWN;","UNLIMITED;   // (" + RemoveEnding(S_dims) + " currently)"),rtnStr,"\r",Inf)
	endif
	
	// get finite dimensions
	NC_ListObjects /Z /Q /T="dim" fileID 
	if (V_flag == 0)
		for (varStr : ListToTextWave(S_names,";"))
			NC_Inquire /Z fileID, varStr
			rtnStr = AddListItem("\t" + varStr + " = " + SelectString(V_flag == 0,"UNKNOWN;",StringFromList(0,S_dims) + ";"),rtnStr,"\r",Inf)
		endfor
	endif

	return rtnStr
End
//___________________________________________________________________________________________________
// get dimensions size
Static Function/S getDim(Wave Wv,String dimStr)

	return RemoveEnding(StringByKey("\t" + dimStr,Note('Wv')," = ","\r"),";")
End
//___________________________________________________________________________________________________
// set the size of a dimension of unknown size i.e. a dimension but not a variable
Static Function setDim(Wave Wv,String dimStr,String dimLen)

	String noteStr = ReplaceString(")\r",Note('Wv'),");\r")
	if (GrepString(getDim(Wv,dimStr),"UNKNOWN"))
		Note /K Wv, ReplaceString(");\r",ReplaceStringByKey("\t" + dimStr,noteStr,dimLen," = ",";\r"),")\r")
	endif
	
	return 0
End
//___________________________________________________________________________________________________
// list the dimesnions 
Static Function/S listDims(Wave Wv)

	String dimStr,rtnStr="" 

	for (String lstStr : ListToTextWave(note(Wv),"\t"))
		SplitString /E=("(.*?) = .*\r") lstStr,dimStr 
		if (strlen(dimStr) > 0)
			rtnStr = AddListItem(dimStr,rtnStr,";",Inf)
		endif
	endfor
	
	return rtnStr
End
//___________________________________________________________________________________________________
// try to determine if a text wave is really char
Static Function isChar(Wave Wv)

	int type = WaveType(Wv)
	
	// is numeric 
	if (type > 0)
		return 0
	endif 
	
	// check string length 
	[int maxlen,double avgLen] = getMaxAvgStrLen(Wv)
	
	if ((maxlen == 1) && (avgLen <= 1))
		return 1
	endif		
		
	return 0
End
//___________________________________________________________________________________________________
// get avgerage string length in a text wave
Static Function [int maxlen,double avgLen] getMaxAvgStrLen(Wave/T Wv)

	Make /D /FREE /N=(numpnts(Wv)) lenWv = strlen(Wv[p])
	WaveStats /Q /Z /M=1 lenWv
	
	return [V_max,V_avg]
End
//___________________________________________________________________________________________________	
// convert a char wave into a text wave
Static Function/WAVE char2text(Wave/T charWv)
	
	double i,j=DimSize(charWv,0)

	Make /O /FREE /T /N=(j) textWv
	
	for (i=0;i<j;i++)
		Duplicate /O /FREE /RMD=[i][] charWV strWv
		textWv[i] = UnPadString(char2Str(strWv),padValue)
	endfor

	return textWv 
End
//___________________________________________________________________________________________________
// convert a char to a string
Static Function/S char2Str(Wave/T charWv)

	String rtnStr
	wfprintf rtnStr,"%s" charWv
	
	return rtnStr
End
//___________________________________________________________________________________________________
// clean up a dimension label of single quotes, double quotes, semi colons and colons
Static Function/S cleanUpDimLabel(String str)

	str = ReplaceString("'",str,"")
	str = ReplaceString("\"",str,"")
	str = ReplaceString(";",str,"")
	str = ReplaceString(":",str,"")

	return str
End
//___________________________________________________________________________________________________
// find out what kind of netCDF we have from full path to file
Static Function netCDFtype(String fNameStr)

	double fileID ; int type
	
	// try netCDF-4
	HDF5OpenFile /Z /R fileID as fNameStr 
	if (V_flag == 0)
		type = 4 + (3 * isClassic(fileID))
		HDF5CloseFile fileID		
		return type
	endif
	
	// try netCDF-3
	NC_OpenFile /Z fileID as fNameStr
	if (V_flag == 0)
		type = 3 * getCDFtype(fNameStr)
		NC_CloseFile fileID
		return type
	endif
	
	// try CDF-5 as a last resort
	type = (5 == getCDFtype(fNameStr)) ? 5 : 0
	
	return type
End
//___________________________________________________________________________________________________
// get the binary header from the netCDF classic file (4th byte either 1, 2 or 5)
Static Function getCDFtype(String ObjectStr)

	DFREF dfrSave = GetDataFolderDFR()

	SetDataFolder NewFreeDataFolder()	// Create new free data folder

	// load the 4th byte of the CDF which contains the type (1, 2 or 5)
	GBLoadWave /T={8,8} /W=1 /S=3 /U=1 /O /Q /A=NCid ObjectStr ; Wave NCid0

	SetDataFolder dfrSave // go home
	
	Return NCid0[Inf] 
End
//___________________________________________________________________________________________________
// get the netCDF type as a string
Static Function/S netCDFtypeStr(int type)

	String rtnStr 
	
	switch (type)
		case 3:
			rtnStr = "netCDF-3"					// classic CDF-1
			break
		case 4:
			rtnStr = "netCDF-4"					// HDF-5
			break
		case 5:
			rtnStr = "CDF-5"					// unicorn
			break
		case 6:
			rtnStr = "netCDF-3 64-bit offset" 	// CDF-2
			break
		case 7:
			rtnStr = "netCDF-4 classic"			// HDF-5
			break
		default:
			rtnstr = "unknown"
	endswitch
	
	return rtnStr
End		
//___________________________________________________________________________________________________	
// test to see if file is netCDF4 'classic model' indicated by the presence of _nc3_strict attribute
// the only valid value of _nc3_strict is 1, no need to check there is no _nc3_strict=0
Static Function isClassic(double fileID)

	// look for the _nc3_strict attribute in the root of the file
	HDF5ListAttributes /Type=1 fileID, "/"

	return GrepString(S_HDF5ListAttributes,"_nc3_strict")
End
//___________________________________________________________________________________________________	
// funtion that does the work of loading netCDF-4 / HDF-5 files
Static Function [int error,String errorStr,String WvListStr] getNC4(String fNameStr,int mode,int transpose)
	
	// initialize the outputs
	error = 0 ; errorStr ="\rLog:\r" ; WvListStr=""

	double fileID,groupID 
	String saveDF = GetDataFolder(1)
	String attStr,varStr,attWvStr,attListStr,varWvStr,varListStr,groupStr,groupListStr
	String funcLstStr = FunctionList("*",";","KIND:7"),tempErrorStr

	// Make a ew symbolic path to the netCDF
	NewPath /O /Q /Z netCDFpath, ParseFilePath(1,fNameStr,":",1,0) ; error = (V_flag || error) ? 1 : 0
	sprintf tempErrorStr, "Errors occurred creating a path: '%s'\r",ParseFilePath(0,fNameStr,":",1,0) ; errorStr += SelectString(V_flag,"",tempErrorStr)
	
	// handle NewPath errors
	if (V_flag)
		return [error,errorStr,WvListStr]
	endif
	
	// open as netCDF4
	HDF5OpenFile /Z /R /P=netCDFpath fileID as fNameStr ; error = (V_flag || error) ? 1 : 0
	sprintf tempErrorStr, "Errors occurred opening: '%s'\r",ParseFilePath(0,fNameStr,":",1,0) ; errorStr += SelectString(V_flag,"",tempErrorStr)

	// handle file open errors 
	if (V_flag)
		return [error,errorStr,WvListStr]
	endif	
	
	// get a list of groups in the file
	HDF5ListGroup /CONT=1 /F /TYPE=1 /R /ORDR=1 /Z fileID,"/" ; groupListStr = S_HDF5ListGroup + "/;" ; error = (V_flag || error) ? 1 : 0
	sprintf tempErrorStr, "Errors occurred listing all groups: '%s'\r",ParseFilePath(0,fNameStr,":",1,0) ; errorStr += SelectString(V_flag,"",tempErrorStr)

	// loop through all the groups in the file
	for (groupStr : ListToTextWave(groupListStr,";"))

		// make a data folder for each group
		if (CmpStr(groupStr,"/",2))	
			NewDataFolder /O /S $ReplaceString("/",PossiblyQuotePath(groupStr,"/"),":")
		endif
				
		// Get a list of attributes for each group
		HDF5ListAttributes /Z /Type=1 fileID,groupStr ; attListStr = S_HDF5ListAttributes ; error = (V_flag || error) ? 1 : 0
		sprintf tempErrorStr, "Errors occurred listing all group attributes: '%s'\r",ParseFilePath(0,fNameStr,":",1,0) ; errorStr += SelectString(V_flag,"",tempErrorStr)

		//strip out HDF5/transparent netCDF markup
		attListStr = RemoveFromList(HDF5markup + NCmarkup,attListStr,";")
		
		// sort attributes alphanumerically
		#ifdef AlphaNumSort
			attListStr = SortList(attListStr,";",16)
		#endif
		
		// Make the info Wave 
		Make /O /T /N=0 $SelectString(CmpStr(groupStr,"/",2),globAttWvStr,grpAttWvStr)/Wave=infoWv
		
		// loop through the global attributes 
		for (attStr : ListToTextWave(attListStr,";"))
			
			// ensure the attribute name won't conflict
			attWvStr = "QLA_temp_"+attStr
			
			// load the attribute
			HDF5LoadData /O /A=attStr /TYPE=1 /N=$attWvStr /OPTS=1 /IGOR=0 /Q /Z fileID, groupStr ; error = (V_flag || error) ? 1 : 0 
			sprintf tempErrorStr, "Errors occurred loading global attribute: '%s'\r",attStr ; errorStr += SelectString(V_flag,"",tempErrorStr)
	
			// error
			if (V_flag)
				QLA_set_att(infoWv,attStr,attErrorStr + ";")
				continue
			// string
			elseif ((WaveType($attWvStr) == 0) && (numpnts($attWvStr) != 0)) 
				Wave /Z /T attStrWv = $attWvStr
				QLA_set_att(infoWv,attStr,"\"" + makeASCII(attStrWv[Inf]) + "\";")
			// numeric
			elseif  ((WaveType($attWvStr) != 0) && (numpnts($attWvStr) != 0))  
				Wave /Z	attVarWv = $attWvStr
				QLA_set_att(infoWv,attStr,cprintf(AttVarWv) + "; // " + TypeToInfo(WaveType(attVarWv)))
			// empty
			else
				QLA_set_att(infoWv,attStr,attNullStr + ";") 
			endif
		
			KillWaves /Z $attWvStr
		// end of global metadata loop
		endfor

		// make the attributes title 
		Note /K infoWv, SelectString(CmpStr(groupStr,"/",2),"// global attributes:\r","// group attributes\r") + note(infoWv)
	
		// Kill the info Wave if it is empty
		if (ItemsInList(QLA_inq_att(infoWv)) == 0)
			KillWaves /Z infoWv
		endif
		
		// Open the group containing the info 
		HDF5OpenGroup /Z fileID, groupStr, groupID ; error = (V_flag || error) ? 1 : 0
		
		// list all the variables in a group
		HDF5ListGroup /Type=2 /ORDR=1 groupID, groupStr ; varListStr = S_HDF5ListGroup  ; error = (V_flag || error) ? 1 : 0
		
		// make a wave to hold the dimension information
		Make /O /T /N=0 $dimWvStr/Wave=dimWv
		Note /k dimWv, "dimensions:\r" + note(dimWv)
		
		// loop though the variables load them and test to see if they are a dimension
		for (varStr : ListToTextWave(varListStr,";"))
		
			// check that the variable name won't conflict
			varWvStr = SelectString(CmpStr(varStr,"time"),timeWvStrConstant,varStr)
			varWvStr = SelectString(WhichListItem(varWvStr,funcLstStr,";",0,0) > 0,varWvStr,illegalPreStr + varStr + illegalSuffStr)
		
			// Load variable
			HDF5LoadData /O /N=$varWvStr /OPTS=0 /IGOR=0 /Q /Z groupID, varStr ; error = (V_flag || error) ? 1 : 0 
			sprintf tempErrorStr, "Errors occurred loading variable: '%s'\r",varStr ; errorStr += SelectString(V_flag,"",tempErrorStr)
		
			// refernece the loaded wave
			Wave dataWv = $varWvStr
			
			// check to see if data was actually loaded 
			if (!WaveExists(dataWv))
				continue
			endif
			
			// add the name to the list of variables 
			WvListStr = AddListItem(varWvStr,WvListStr,";",Inf)
			
			// kill the wavenote set the short name
			Note /K dataWv ; QLA_set_att(dataWv,"short_name","\"" + varStr + "\";")
			
			// Get a list of attributes for each variable
			HDF5ListAttributes /Type=2 groupID, varStr ; attListStr = S_HDF5ListAttributes ; error = (V_flag || error) ? 1 : 0 
			sprintf tempErrorStr, "Errors occurred listing attributes of variable: '%s'\r",varStr ; errorStr += SelectString(V_flag,"",tempErrorStr)
	
			// strip out HDF-5 and netCDF hidden attributes
			attListStr = RemoveFromList(HDF5markup + NCmarkup,attListStr,";")
			
			// sort attributes alphanumerically
			#ifdef AlphaNumSort
				attListStr = SortList(attListStr,";",16)
			#endif
			
			// loop through the variable attributes 
			for (attStr : ListToTextWave(attListStr,";"))
			
				// ensure the attribute name won't conflict
				attWvStr = "QLA_temp_"+attStr
			
				// load the attribute
				HDF5LoadData /O /A=attStr /TYPE=2 /N=$attWvStr /OPTS=1 /IGOR=0 /Q /Z groupID, varStr ; error = (V_flag || error) ? 1 : 0 
				sprintf tempErrorStr, "Errors occurred loading the attribute: '%s' of variable: '%s'\r", attStr, varStr ; errorStr += SelectString(V_flag,"",tempErrorStr)

				// error
				if (V_flag)
					QLA_set_att(dataWv,attStr,attErrorStr + ";")
					continue
				// string
				elseif ((WaveType($attWvStr) == 0) && (numpnts($attWvStr) != 0)) 
					Wave /Z /T attStrWv = $attWvStr
					QLA_set_att(dataWv,attStr,"\"" + makeASCII(attStrWv[Inf]) + "\";")
				// numeric
				elseif ((WaveType($attWvStr) != 0) && (numpnts($attWvStr) != 0))  
					Wave /Z	attVarWv = $attWvStr
					QLA_set_att(dataWv,attStr,cprintf(attVarWv) + "; // " + TypeToInfo(WaveType(attVarWv)))
				// empty
				else
					QLA_set_att(dataWv,attStr,attNullStr + ";") 
				endif
				
				KillWaves /Z $attWvStr
			// end of variable metadata loop
			endfor 
		
			// check to see if the variable is also an attribute
			Note /k dimWv,note(dimWv) + isHDF5Dimension(groupID,varStr)
		
			// apply metadata conventions to wave
			error += unpackNC(dataWv,mode)
		
			// set the dimensional dependencies
			error += setDependencies(dataWv,getHDF5Dimensions(groupID,varStr))
			
			// set the scaling of time if it exists
			if (WaveExists($timeWvStrConstant))
				SetScale /P x 0,1,QLA_get_att($timeWvStrConstant,"units"), $timeWvStrConstant
				SetScale d 0,0,"dat", $timeWvStrConstant
			endif
		
			// convert char to text if required
			#ifdef CharAsText
				if (isCHar(dataWv))
					Wave textWv = char2text(dataWv)
					Note /K textWv, note(dataWv)
					Duplicate /O textWv dataWv
				endif				
			#endif
			
			// transpose 2D datasets if required
			if (transpose && (WaveDims(dataWv) > 1))
				 MatrixTranspose dataWv
			endif
	
		// end of variable loop
		endfor
	
		// no dimensions defined in this group
		if (ItemsInList(listDims(dimWv)) == 0)
			KillWaves dimWv
		endif

		SetDataFolder saveDF
	
	// end of group loop
	endfor
	
	// Close the file
	HDF5CloseFile /Z fileID ; error = (V_flag || error) ? 1 : 0 
	sprintf tempErrorStr, "Errors occurred closing: '%s'\r",ParseFilePath(0,fNameStr,":",1,0) ; errorStr += SelectString(V_flag,"",tempErrorStr)

	// Retrieve collected error information
//	HDF5DumpErrors /CLR=1 ; Print S_HDF5ErrorInfo
	
	// clean up
//	KillPath /Z netCDFpath
	
	return [error,errorStr,WvListStr]
End
//___________________________________________________________________________________________________	
// funtion that does the work of loading netCDF-3 files
Static Function [int error,String errorStr,String WvListStr] getNC3(String fNameStr,int mode,int transpose)

	// initialize the outputs
	error = 0 ; errorStr="\rLog:\r" ; WvListStr="" 

	double fileID
	String saveDF = GetDataFolder(1)
	String attStr,varStr,attWvStr,attListStr,varWvStr,varListStr
	String funcLstStr = FunctionList("*",";","KIND:7"),tempErrorStr

	// Make a new symbolic path to the netCDF
	NewPath /O /Q /Z netCDFpath, ParseFilePath(1,fNameStr,":",1,0) ; error = (error || V_flag) ? 1 : 0
	sprintf tempErrorStr, "Errors occurred creating a path: '%s'\r",ParseFilePath(0,fNameStr,":",1,0) ; errorStr += SelectString(V_flag,"",tempErrorStr)

	// handle NewPath errors
	if (V_flag)
		return [error,errorStr,WvListStr]
	endif

	// Open the netCDF-3				
	NC_OpenFile /Z /PATH=netCDFpath fileID as fNameStr ; error = (V_flag || error) ? 1 : 0
	sprintf tempErrorStr, "Errors occurred opening: '%s'\r",ParseFilePath(0,fNameStr,":",1,0) ; errorStr += SelectString(V_flag,"",tempErrorStr)
	
	// handle file open errors 
	if (V_flag)
		return [error,errorStr,WvListStr]
	endif	

	// Get global attributes
	Make /O /T /N=0 $globAttWvStr/Wave=infoWv

	// list all the global attributes 
	NC_ListObjects /Z /Q /T="att" fileID ; attListStr = S_names ; error = (V_flag || error) ? 1 : 0
	sprintf tempErrorStr, "Errors occurred listing all global attributes: '%s'\r",ParseFilePath(0,fNameStr,":",1,0) ; errorStr += SelectString(V_flag,"",tempErrorStr)

	// sort attributes alphanumerically
	#ifdef AlphaNumSort
		globAttListStr = SortList(globAttListStr,";",16)
	#endif		 			

	// loop through the global attributes 
	for (attStr : ListToTextWave(attListStr,";"))
	
		// ensure the attribute name won't conflict
		attWvStr = "QLA_temp_"+attStr
	
		// load the attribute
		NC_LoadData /Z /Q /O /T=1 /DEST=$attWvStr /A=attStr fileID ; error = (V_flag || error) ? 1 : 0
		sprintf tempErrorStr, "Errors occurred loading global attribute: '%s'\r",attStr ; errorStr += SelectString(V_flag,"",tempErrorStr)

		// error
		if (V_flag)
			QLA_set_att(infoWv,attStr,attErrorStr + ";")
			continue
		// string
		elseif ((WaveType($attWvStr) == 0) && (numpnts($attWvStr) != 0)) 
			Wave /Z /T attStrWv = $attWvStr
			QLA_set_att(infoWv,attStr,"\"" + makeASCII(attStrWv[Inf]) + "\";")
		// numeric
		elseif  ((WaveType($attWvStr) != 0) && (numpnts($attWvStr) != 0))  
			Wave /Z	attVarWv = $attWvStr
			QLA_set_att(infoWv,attStr,cprintf(attVarWv) + "; // " + TypeToInfo(WaveType(attVarWv)))
		// empty
		else
			QLA_set_att(infoWv,attStr,attNullStr + ";") 
		endif
		
		KillWaves /Z $attWvStr
	// end of global metadata loop
	endfor 
	
	// make the attributes title 
	Note /K infoWv, "// global attributes:\r" + note(infoWv)
	
	// Kill the info Wave if it is empty
	if (ItemsInList(QLA_inq_att(infoWv)) == 0)
		KillWaves /Z infoWv
	endif
		
	// get dimensions
	Make /O /T /N=0 $dimWvStr/Wave=dimWv
	Note dimWv,getNCdimensionStr(fileID)
		
	// list all variables in file
	NC_ListObjects /Z /T="var" fileID ; varListStr = S_names ; error = (V_flag || error) ? 1 : 0
	sprintf tempErrorStr, "Errors occurred listing all variables: '%s'\r",ParseFilePath(0,fNameStr,":",1,0) ; errorStr += SelectString(V_flag,"",tempErrorStr)
	
	// Get the variables	
	for (varStr : ListToTextWave(varListStr,";"))
		
		// check that the variable name won't conflict
		varWvStr = SelectString(CmpStr(varStr,"time"),timeWvStrConstant,varStr)
		varWvStr = SelectString(WhichListItem(varWvStr,funcLstStr,";",0,0) > 0,varWvStr,illegalPreStr + varStr + illegalSuffStr)
			
		// Load variable
		NC_LoadData /Z /O /U=3 /T=0 /DEST=$varWvStr fileID, varStr ; error = (V_flag || error) ? 1 : 0
		sprintf tempErrorStr, "Errors occurred loading variable: '%s'\r",varStr ; errorStr += SelectString(V_flag,"",tempErrorStr)
		
		// refernece the loaded wave
		Wave dataWv = $varWvStr
		
		// check to see if data was actually loaded 
		if (!WaveExists(dataWv))
			continue
		endif
		
		// add the name to the list of variables 
		WvListStr = AddListItem(varWvStr,WvListStr,";",Inf)
		
		// kill the wavenote set the short name
		Note /K dataWv ; QLA_set_att(dataWv,"short_name","\"" + varStr + "\";")
		
		// list all variables for the attribute
		NC_ListAttributes /Z /Q fileID, varStr ; attListStr = S_names ; error = (V_flag || error) ? 1 : 0
		sprintf tempErrorStr, "Error occurred listing the attributes of variable: '%s'\r",varStr ; errorStr += SelectString(V_flag,"",tempErrorStr)

		// sort attributes alphanumerically
		#ifdef AlphaNumSort
			attListStr = SortList(attListStr,";",16)
		#endif
				
		// loop through the variable attributes 
		for (attStr : ListToTextWave(attListStr,";"))
		
			// ensure the attribute name won't conflict
			attWvStr = "QLA_temp_"+attStr
		
			// loat the attribute
			NC_LoadData /Z /Q /O /T=1 /DEST=$attWvStr /A=attStr fileID, varStr ; error = (V_flag || error) ? 1 : 0
			sprintf tempErrorStr, "Errors occurred loading the attribute: '%s' of variable: '%s'\r", attStr,varStr ; errorStr += SelectString(V_flag,"",tempErrorStr)
	
			// error
			if (V_flag)
				QLA_set_att(dataWv,attStr,attErrorStr + ";")
				continue
			// string
			elseif ((WaveType($attWvStr) == 0) && (numpnts($attWvStr) != 0)) 
				Wave /Z /T attStrWv = $attWvStr
				QLA_set_att(dataWv,attStr,"\"" + makeASCII(attStrWv[Inf]) + "\";")
			// numeric
			elseif  ((WaveType($attWvStr) != 0) && (numpnts($attWvStr) != 0))  
				Wave /Z	attVarWv = $attWvStr
				QLA_set_att(dataWv,attStr,cprintf(attVarWv) + "; // " + TypeToInfo(WaveType(attVarWv)))
			// empty
			else
				QLA_set_att(dataWv,attStr,attNullStr + ";") 
			endif
			
			KillWaves /Z $attWvStr
		// end of variable metadata loop
		endfor 
		
		// apply metadata conventions to wave
		error += unpackNC(dataWv,mode)
			
		// get dimension information
		NC_Inquire /Z fileID, varStr ; error = (V_flag || error) ? 1 : 0
		sprintf tempErrorStr, "Errors occurred getting dimension information of variable: '%s'\r",varStr ; errorStr += SelectString(V_flag,"",tempErrorStr)
		error += setDependencies(dataWv,S_names)
		
		// deal with the size of annonymous dimensions
		error += setDim(dimWv,ListToTextWave(S_names,";")[Inf],ListToTextWave(S_dims,";")[Inf])

		// convert char to text if required
		#ifdef CharAsText
			if (isCHar(dataWv))
				Wave textWv = char2text(dataWv)
				Note /K textWv, note(dataWv)
				Duplicate /O textWv dataWv
			endif				
		#endif
		
		// transpose 2D datasets if required
		if (transpose && (WaveDims(dataWv) > 1))
			 MatrixTranspose dataWv
		endif

	// end of variable metadata loop
	endfor
		
	// set the scaling of time if it exists
	if (WaveExists($timeWvStrConstant))
		SetScale /P x 0,1,QLA_get_att($timeWvStrConstant,"units"), $timeWvStrConstant
		SetScale d 0,0,"dat", $timeWvStrConstant
	endif

	// close the file
	NC_CloseFile /Z fileID ; error = (V_flag || error) ? 1 : 0
	sprintf tempErrorStr, "Errors occurred closing: '%s'\r",ParseFilePath(0,fNameStr,":",1,0) ; errorStr += SelectString(V_flag,"",tempErrorStr)

	// Retrieve collected error information
//	NC_DumpErrors /C ; Print S_errorInfo

	// clean up
//	KillPath /Z netCDFpath
	
	return  [error,errorStr,WvListStr]
End
//___________________________________________________________________________________________________	
// function to create and display a dump of the netCDF header CDL text 
Function QLA_dump()
	
	String pathStr,nbName,ncInfoWvLst,ncInfoWvStr ; int iter
	DFREF saveDFR = GetDataFolderDFR()	// get the current data folder
	
	// get a lis of all 'file_attributes' waves relative to the current DF
	ncInfoWvLst = listWaves(saveDFR,1,matchStr=fileAttWvStr)

	// loop though all the folders containing file_attributes waves
	for (ncInfoWvStr : ListToTextWave(ncInfoWvLst,";"))
		pathStr = ParseFilePath(1,ncInfoWvStr,":",1,0)
	 	nbName = QLA_get_att($ncInfoWvStr,"filename")
	 	DoWindow/K $num2str(iter,"nc_info_%d") 
		iter += getCDLnote($pathStr,0,nbName,iter)
		print2notebook("}",NBName,iter-1)
	endfor

	return 0
End
//___________________________________________________________________________________________________
// function to extract the CDL formatted wavenote and structure it correctly 
Static Function getCDLnote(DFREF dfr,int level,String winStr,int iter)
	
	int i,j ; String name,cmdStr,grepStr
	winStr = ReplaceString("'",winStr,"") // unquote name
	
	// get the indent for the current tree location
	String indentStr = ReplicateString("\t",level+1)
	
	// first iteration
	if (level == 0)
		name = GetDataFolder(1, dfr)
		sprintf cmdStr,"netcdf file: %s {\r",winStr
		print2notebook(cmdStr,winStr,iter)
	endif
		
	// dimensions
	int dimWaves = CountObjectsDFR(dfr, 1)
	for(i=0;i<dimWaves;i++)
		name = GetIndexedObjNameDFR(dfr, 1, i)
		sprintf grepStr,"%s",dimWvStr
		if (GrepString(name,grepStr))
			wave Wv = WaveRefIndexedDFR(dfr,i)
			sprintf cmdStr, "%s%s\r", indentStr, ReplaceString("\r\t",makeASCII(note(wv)),"\r\t"+indentStr)
			print2notebook(cmdStr,winStr,iter)
		endif
	endfor 
	
	// variables
	int varWaves = CountObjectsDFR(dfr, 1)
	for (i=0;i<varWaves;i++)
		name = GetIndexedObjNameDFR(dfr, 1, i)
		sprintf grepStr,"%s|%s|%s|%s",globAttWvStr,grpAttWvStr,fileAttWvStr,dimWvStr
		if (!GrepString(name,grepStr))
			
			// print the section opener if this is the first variable
			cmdStr = SelectString(j,indentStr+"variables:\r","") ; j++ ; print2notebook(cmdStr,winStr,iter)
			
			// get the note of the variable and format it correctly
			Wave Wv = WaveRefIndexedDFR(dfr,i)
			sprintf cmdStr, "%s\t%s \r", indentStr,alignTabs(makeASCII(note(wv)),"\t\t"+indentStr)
			print2notebook(cmdStr,winStr,iter)
		endif
	endfor
	
	// groups
	int numDataFolders = CountObjectsDFR(dfr, 4)  
	for (i=0;i<numDataFolders;i++)
		name = GetIndexedObjNameDFR(dfr, 4, i)
		
		// open group
		sprintf cmdStr, "%sgroup: %s {\r", indentStr, name
		print2notebook(cmdStr,winStr,iter)
		
		// recurse
		DFREF childDFR = dfr:$(name)
		getCDLnote(childDFR,level+1,winStr,iter)
		
		// close group
		sprintf cmdStr, "%s }\r\r", indentStr
		print2notebook(cmdStr,winStr,iter)
	endfor 
	
	// group attributes
	int grpAttWaves = CountObjectsDFR(dfr, 1)
	for (i=0;i<grpAttWaves;i++)
		name = GetIndexedObjNameDFR(dfr, 1, i)
		sprintf grepStr,"%s",grpAttWvStr
		if (GrepString(name,grepStr))
			Wave Wv = WaveRefIndexedDFR(dfr,i)
			sprintf cmdStr, "%s%s \r", indentStr,alignTabs(makeASCII(note(wv)),indentStr) 
			print2notebook(cmdStr,winStr,iter)
		endif
	endfor 
	
	// global attributes
	int globAttWaves = CountObjectsDFR(dfr, 1)
	for (i=0;i<globAttWaves;i++)
		name = GetIndexedObjNameDFR(dfr, 1, i)
		sprintf grepStr,"%s",globAttWvStr
		if (GrepString(name,grepStr))
			Wave Wv = WaveRefIndexedDFR(dfr,i)
			sprintf cmdStr, "%s%s \r", indentStr,alignTabs(makeASCII(note(wv)),indentStr)
			print2notebook(cmdStr,winStr,iter)
		endif
	endfor
	
	return 1
End
//___________________________________________________________________________________________________
// function to add text to the nc_info_X notebook
Static Function print2notebook(String infoStr,String winNameStr,int iter)
	
	// get a unique window name
	String infoWinID = num2str(iter,"nc_info_%d"),cmdStr
	
	// check if the window already exists	
	DoWindow $infoWinID
	if(V_flag != 1) // window does not exist so create it
		NewNoteBook/F=0/K=1 /N=$infoWinID /V=1 as winNameStr
	endif	
	
	// bring it to front
	DoWindow/F $infoWinID
	
	// go to the end of the text
	Notebook $infoWinID selection={endOfFile,endOfFile}
	
	// put the text in the notebook
	Notebook $infoWinID text=infoStr

	// got to the top of the notebook
	Notebook $infoWinID selection={startOfFile,startOfFile}, findText={"",1}
	
	return 0
End
//___________________________________________________________________________________________________
// get comments aligned properly
Static Function/S alignTabs(String str,String indentStr)

	str = ReplaceString("\t",str,"")
	str = ReplaceString("\r",str,"\r\t"+indentStr)
	str = ReplaceString("\t:",str,":")
	str = ReplaceString("\n",str,"\r\t"+indentStr)
	str = ReplaceString("\r\n",str,"\r\t"+indentStr)
	
	return str
End
//___________________________________________________________________________________________________
// function to find get paths to all waves, or just those matching matchStr
Static Function/S listWaves(DFREF dfr,int recurse,[,String matchStr])

	int i,numWv,numDF ; String name,path,rtnStr=""
	
	// match everyting by default
	matchStr = SelectString(ParamIsDefault(matchStr),matchStr,"*")

	// loop through all the waves in the data folder
	numWv = CountObjectsDFR(dfr, 1)
	for(i=0;i<numWv;i++)
		Wave /Z Wv = WaveRefIndexedDFR(dfr,i)
		name = NameOfWave(Wv)
		path = GetWavesDataFolder(Wv, 2)
		rtnStr += SelectString(StringMatch(name,matchStr),"",path + ";")
	endfor

	// loop through all data folders recursively
	if (recurse)
		numDF = CountObjectsDFR(dfr, 4)  
		for(i=0;i<numDF;i++)
			name = GetIndexedObjNameDFR(dfr, 4, i)
			DFREF childDFR = dfr:$(name)
			rtnStr += listWaves(childDFR,recurse,matchStr=matchStr)
		endfor 
	endif
	
	return rtnStr
End
//___________________________________________________________________________________________________
// return the datatype of a variable in a HDF5 file
Static Function typeOnDisk(String path,double fileID)
	
	STRUCT HDF5DataInfo di	
		
	InitHDF5DataInfo(di)	
	int dummy = HDF5DatasetInfo(fileID, path, 0, di)
	int type = h5DataType2WaveType(di.datatype_str)
	
	// check for variable length strings to differentiate CHAR
	type = (type == 0) ? type + (!di.datatype_is_vlen) : type 

	return type
End
//___________________________________________________________________________________________________
// get wavetype from HDF5 data type string
Static Function h5DataType2WaveType(String h5DataType)

	int type 
	
	strswitch (h5DataType)
		case "string":
			type = 0	// string
			break
		case "32-bit LE float":
			type = 2	// float / real
			break
		case "64-bit LE float":
			type = 4	// double
			break
		case "8-bit signed LE integer":
			type = 8	// byte
			break
		case "16-bit signed LE integer":
			type = 16	// short
			break
		case "32-bit signed LE integer":
			type = 32	// long / int
			break
		case "8-bit unsigned LE integer":
			type = 72	// ubyte
			break
		case "16-bit unsigned LE integer":
			type = 80	// ushort
			break
		case "32-bit unsigned LE integer":
			type = 96	// uint
			break
		case "64-bit signed LE integer":
			type = 128	// int64 
			break
		case "64-bit unsigned LE integer":
			type = 192	// uint64 
			break
		default:
			type = 4 	// cover new cases
	endswitch
	
	return type
End 
//___________________________________________________________________________________________________
// Get the netCDF type returned as a string
Function/S QLA_kind(String fNameStr)
	
	double fileID
	
	// check that a valid full path was given
	GetFileFolderInfo /Q /Z fNameStr
	
	// prompt for a file if invalid path given
	if (V_flag != 0)
		Open /D /R /F=fileFilters /M="Choose a netCDF" fileID ; fNameStr = S_filename 
	endif
	
	return netCDFtypeStr(netCDFtype(fNameStr))
End
//___________________________________________________________________________________________________
// function to nix all non-ASCII characters from a string - from Adam Light WM
Static Function/S makeASCII(String inStr)
	
	// Store the bytes of the string in a byte wave so we can
	// use wave related commands.
	Wave /B /U byteWv = StringToUnsignedByteWave(inStr)
	
	// Set all non-printable characters to 0 (null)
	// The test below considers tab (9), line feed (10), and carriage return (13)
	// to be printable characters.
	MultiThread byteWv = (byteWv[p] >= 32 && byteWv[p] <= 126) || (byteWv[p] == 9 || byteWv[p] == 10 || byteWv[p] == 13) ? byteWv[p] : 0
	
	// Remove all of the nulls from the wave
	WaveTransform zapZeros byteWv        
	
	// Convert the bytes in the wave back into a string
	String outStr = WaveDataToString(byteWv)  
	
	return outStr
End
//___________________________________________________________________________________________________

